"""define loader objects that return PowderDiffraction Data Structures"""
import io
import math

import numpy as np
import pandas as pd

from radie import exceptions
from radie.loaders import Loader, register_loaders
from radie.plugins.structures.powderdiffraction import PowderDiffraction, CuKa, CuKa1, CuKa2


def ras_meta_value(line: bytes, return_type: type=str):
    value = line[line.find(b" "):].strip().replace(b'"', b"")
    if return_type is str:
        return value.decode("ascii")
    else:
        return return_type(value)


def asc_meta_value(line: str, return_type: type=str):
    val = line.split("=")[1].strip()
    if return_type is str:
        return val
    else:
        return return_type(val)


def load_asc(fname, name=None):
    """
    .ras file output from Rigaku XRD.  Tested with files from MiniFlex system, which seem to be bytes-like

    Parameters
    ----------
    fname : str
        filename
    name : str
        measurement identifier

    Returns
    -------
    PowderDiffraction

    """

    with open(fname, "r") as fid:
        lines = fid.readlines()

    # --- begin file check --- #
    if not all((
        lines[0].startswith("*TYPE"),
        lines[8].startswith("*GONIO"),
        lines[41].startswith("*BEGIN"),
        lines[77].startswith("*COUNT"),
    )):
        exceptions.IncorrectFileType

    wavelength1 = asc_meta_value(lines[23], float)
    wavelength2 = asc_meta_value(lines[24], float)
    if wavelength1 == 1.54059 and wavelength2 == 1.54441:
        wavelength = CuKa
    else:
        wavelength = wavelength1
    sample_name = asc_meta_value(lines[2], str)  # type: str
    start = asc_meta_value(lines[43], float)  # type: float
    stop = asc_meta_value(lines[44], float)  # type: float
    num_points = asc_meta_value(lines[77], int)  # type: int
    source = asc_meta_value(lines[23], str)  # type: str

    twotheta = np.linspace(start, stop, num_points, dtype=float)
    data_lines = int(math.ceil(num_points / 4))
    datastart = 78
    intensities = list()
    try:
        for i in range(datastart, datastart + data_lines):
            for point in map(float, lines[i].split(", ")):
                intensities.append(point)
    except Exception:
        raise exceptions.LoaderException

    return PowderDiffraction(
        data=np.array((twotheta, intensities), dtype=float).T, columns=("twotheta", "intensity"),
        name=sample_name, wavelength=wavelength, source=source
    )


def load_ras(fname, name=None):
    """
    .ras file output from Rigaku XRD.  Tested with files from MiniFlex system, which seem to be bytes-like

    Parameters
    ----------
    fname : str
        filename
    name : str
        measurement identifier

    Returns
    -------
    df_xrd : PowderDiffraction
        PowderDiffraction StructuredDataFrame based on XRD data

    """

    with open(fname, "rb") as fid:
        lines = fid.readlines()

    wavelength = None  # type: float
    wavelength1 = None  # type: float
    wavelength2 = None  # type: float
    sample_name = "xray diffaction data"
    start = None  # type: float
    stop = None  # type: float
    step = None  # type: int
    data_start = 0

    for i, line in enumerate(lines):
        if line.startswith(b"*HW_XG_WAVE_LENGTH_ALPHA1 "):
            wavelength1 = ras_meta_value(line, float)
        elif line.startswith(b"*HW_XG_WAVE_LENGTH_ALPHA2 "):
            wavelength2 = ras_meta_value(line, float)
        elif line.startswith(b"*FILE_SAMPLE "):
            sample_name = ras_meta_value(line, str)
        elif line.startswith(b"*MEAS_SCAN_STEP "):
            step = ras_meta_value(line, float)
        elif line.startswith(b"*MEAS_SCAN_START "):
            start = ras_meta_value(line, float)
        elif line.startswith(b"*MEAS_SCAN_STOP "):
            stop = ras_meta_value(line, float)
        elif line.startswith(b"*RAS_INT_START"):
            data_start = i + 1
            break

    if (
        wavelength1 is None or
        start is None or
        stop is None or
        step is None or
        data_start == 0
    ):
        raise exceptions.IncorrectFileType

    if wavelength1 == 1.540593 and wavelength2 == 1.544414:
        wavelength = CuKa
    else:
        wavelength = wavelength1

    n_points = int((stop - start) / step + 1)
    data_stop = data_start + n_points
    data_lines = [line.strip().decode("ascii") for line in lines[data_start:data_stop]]

    data = "\n".join(data_lines)

    if not name:
        name = sample_name

    data_buff = io.StringIO("twotheta intensity uncertainty\n" + data)
    df_xrd = PowderDiffraction(pd.read_csv(data_buff, sep=" "),
                               name=name,
                               wavelength=wavelength,
                               source="CuKa",
                               xunit="deg",
                               yunit="counts")

    return df_xrd


rigaku_ras_loader = Loader(load_ras, PowderDiffraction, [".ras"], "Rigaku XRD")
rigaku_asc_loader = Loader(load_asc, PowderDiffraction, [".asc"], "Rigaku XRD")

register_loaders(
    rigaku_ras_loader,
    rigaku_asc_loader,
)



import sys

from PyQt5 import QtCore, QtWidgets, QtGui
import sip
import pyqtgraph as pg

from ..structures import StructuredDataFrame
from ..import loaders
from . import cfg, visualizations, dpi, masterdftree
from . import functions as fn

try:
    from . import excel_exporter
except ImportError:
    pass


class SubWindow(QtWidgets.QMdiSubWindow):
    pass


class VisualizationWindow(SubWindow):
    def __init__(self, visualization: visualizations.base.Visualization=None, parent=None):
        super(VisualizationWindow, self).__init__(parent)

        if visualization:
            self.setWidget(visualization)

    def setWidget(self, visualization: visualizations.base.Visualization):
        if not isinstance(visualization, visualizations.base.Visualization):
            raise TypeError("Visualization Windows can only have Visualization Widgets as their main widget")
        self.resize(visualization.size())
        super(VisualizationWindow, self).setWidget(visualization)
        self.setWindowIcon(visualization.icon())


class SubWindowListItem(QtWidgets.QListWidgetItem):
    def __init__(self, subwindow: SubWindow, name: str=None, parent=None):
        super(SubWindowListItem, self).__init__(parent)
        self.subWindow = subwindow  # type: QtWidgets.QMdiSubWindow
        self.setFlags(self.flags() | QtCore.Qt.ItemIsEditable)
        if not name:
            self.setText("sub window")
        else:
            self.setText(name)
        self.setIcon(self.subWindow.widget().icon())

    def showVisualization(self):
        if self.subWindow.isHidden():
            self.subWindow.show()
            self.subWindow.widget().show()

        if self.subWindow.isMinimized():
            self.subWindow.setWindowState(QtCore.Qt.WindowNoState)

        self.subWindow.mdiArea().setActiveSubWindow(self.subWindow)

    def setText(self, atext: str):
        super(SubWindowListItem, self).setText(atext)
        self.subWindow.setWindowTitle(atext)


class SubWindowList(QtWidgets.QListWidget):
    visualizationMenuRequested = QtCore.pyqtSignal(object)

    def __init__(self, *args, **kwargs):
        super(SubWindowList, self).__init__(*args, **kwargs)
        self.setAcceptDrops(True)
        self.setDragDropMode(QtWidgets.QAbstractItemView.InternalMove)
        self.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
        self.setEditTriggers(QtWidgets.QAbstractItemView.SelectedClicked | QtWidgets.QAbstractItemView.EditKeyPressed)
        self.itemDoubleClicked.connect(self.processDoubleClick)
        self.itemDelegate().commitData.connect(self.itemTextUpdated)
        self.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self.rightClicked)

    def itemTextUpdated(self):
        item = self.currentItem()  # type: SubWindowListItem
        item.subWindow.setWindowTitle(item.text())

    def processDoubleClick(self, item: SubWindowListItem):
        item.showVisualization()

    def rightClicked(self, pos: QtCore.QPoint):
        selected = self.selectedItems()
        num_rows = len(selected)

        if num_rows == 0:
            self.visualizationMenuRequested.emit(self.mapToGlobal(pos))
            return

        # else we have selected rows, so do something with those rows
        menu = QtWidgets.QMenu()
        if num_rows == 1:
            deleteItem = QtWidgets.QAction("Delete Visualization", menu)
        else:
            deleteItem = QtWidgets.QAction("Delete Visualizations", menu)

        deleteItem.triggered.connect(self.deleteSelectedVisualizations)
        menu.addAction(deleteItem)
        menu.exec_(self.mapToGlobal(pos))  # QtWidgets.QAction

    def deleteSelectedVisualizations(self):
        for item in self.selectedItems():  # type: SubWindowListItem
            item.subWindow.mdiArea().removeSubWindow(item.subWindow)
            sip.delete(item.subWindow)  # necessary to completely remove the QWidgets from memory
            self.takeItem(self.row(item))


class NewVisAction(QtWidgets.QAction):
    newVisualizationRequested = QtCore.pyqtSignal(object)

    def __init__(self, *args, **kwargs):
        super(NewVisAction, self).__init__(*args, **kwargs)
        self.vis_class = None  # class: visualizations.base.Visualization
        self.triggered.connect(self.requestNewVisualization)

    def requestNewVisualization(self, checked: bool):
        self.newVisualizationRequested.emit(self.vis_class)


class VisualizationsList(SubWindowList):
    def __init__(self, parent=None):
        super(VisualizationsList, self).__init__(parent)


class MainWindow(QtWidgets.QMainWindow):

    def __init__(self, parent=None):
        super(MainWindow, self).__init__(parent)

        self.setWindowTitle("Radie")
        self.setWindowIcon(QtGui.QIcon(cfg.radie_icon))

        self.resize(*dpi.width_by_height(1024, 600))
        self.centralwidget = QtWidgets.QWidget(self)
        self.verticalLayout_centralWidget = QtWidgets.QVBoxLayout(self.centralwidget)
        self.verticalLayout_centralWidget.setContentsMargins(0, 0, 0, 0)
        self.mdiArea = QtWidgets.QMdiArea(self.centralwidget)
        self.verticalLayout_centralWidget.addWidget(self.mdiArea)
        self.setCentralWidget(self.centralwidget)

        self.dock_dataFrames = QtWidgets.QDockWidget(self)
        self.dock_dataFrames.setWindowTitle("DataFrames")
        self.treeView_dataFrames = masterdftree.DFTreeView()
        self.dock_dataFrames.setWidget(self.treeView_dataFrames)
        self.addDockWidget(QtCore.Qt.DockWidgetArea(1), self.dock_dataFrames)

        self.dock_visualizations = QtWidgets.QDockWidget(self)
        self.dock_visualizations.setWindowTitle("Visualizations")
        self.visList = VisualizationsList()
        self.visList.visualizationMenuRequested.connect(self.showNewVisualizationsMenu)
        self.dock_visualizations.setWidget(self.visList)
        self.addDockWidget(QtCore.Qt.DockWidgetArea(1), self.dock_visualizations)
        self.setAcceptDrops(True)

        # --- MainWindow Menu Bar Setup ---#
        self.menubar = QtWidgets.QMenuBar(self)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 1024, 26))

        self.menuFile = QtWidgets.QMenu("&File", self.menubar)
        self.actionExit = QtWidgets.QAction("&Exit", self)
        self.actionExit.triggered.connect(self.close)
        self.menuFile.addAction(self.actionExit)
        self.actionSave = QtWidgets.QAction("&Save", self)
        self.menuFile.addAction(self.actionSave)
        self.menuFile.addSeparator()
        self.action_importFromClipboard = QtWidgets.QAction("Import from &Clipboard", self)
        self.action_importFromClipboard.triggered.connect(self.importDataFrameFromClipboard)
        self.menuFile.addAction(self.action_importFromClipboard)
        self.menubar.addAction(self.menuFile.menuAction())

        self.menuEdit = QtWidgets.QMenu("&Edit", self.menubar)
        self.actionCopy = QtWidgets.QAction("&Copy", self)
        self.menuEdit.addAction(self.actionCopy)
        self.actionPaste = QtWidgets.QAction("&Paste", self)
        self.menuEdit.addAction(self.actionPaste)
        self.menubar.addAction(self.menuEdit.menuAction())

        self.menuView = QtWidgets.QMenu("&View", self.menubar)
        self.actionData_Sets = QtWidgets.QAction("&Data Sets", self)
        self.actionData_Sets.triggered.connect(self.dock_dataFrames.show)
        self.menuView.addAction(self.actionData_Sets)
        self.actionVisualizations = QtWidgets.QAction("&Visualizations", self)
        self.actionVisualizations.triggered.connect(self.dock_visualizations.show)
        self.menuView.addAction(self.actionVisualizations)
        self.menubar.addAction(self.menuView.menuAction())

        self.menuVisualizations = QtWidgets.QMenu("V&isualizations", self.menubar)
        self.menubar.addAction(self.menuVisualizations.menuAction())

        self.menuHelp = QtWidgets.QMenu("&Help", self.menubar)
        self.menubar.addAction(self.menuHelp.menuAction())

        self.setMenuBar(self.menubar)

        self.statusbar = QtWidgets.QStatusBar(self)
        self.setStatusBar(self.statusbar)

        self.actionSave.setIcon(fn.icon("saveicon.svg"))
        self.actionCopy.setIcon(fn.icon("copyicon.svg"))
        self.actionPaste.setIcon(fn.icon("pasteicon.svg"))
        self.actionExit.setIcon(fn.icon("exiticon.svg"))
        # --- End MainWindow Menu Bar Setup --- #

        ## initialize visualization menu items
        self.vis_actions = []
        for cls in visualizations.visualizations.values():  # type : visualizations.base.Visualization
            action = NewVisAction("add &" + cls.name, self)
            action.vis_class = cls
            action.setIcon(cls.icon())
            action.newVisualizationRequested.connect(self.addNewVisualization)
            self.vis_actions.append(action)
            self.menuVisualizations.addAction(action)

        self.menubar.addAction(self.menuVisualizations.menuAction())
        ## end menus

        self.actionSave.setEnabled(False)

    def showNewVisualizationsMenu(self, pos: QtCore.QPoint):
        menu = QtWidgets.QMenu()
        for action in self.vis_actions:
            menu.addAction(action)
        menu.exec_(pos)

    def closeEvent(self, event):
        quit_msg = ""

        quit_msg += "Are you sure you want to exit the program?"
        reply = QtWidgets.QMessageBox.question(self, 'Message',
                                               quit_msg, QtWidgets.QMessageBox.Yes, QtWidgets.QMessageBox.No)

        if reply == QtWidgets.QMessageBox.Yes:
            event.accept()
        else:
            event.ignore()

    def dragEnterEvent(self, a0: QtGui.QDragEnterEvent):
        if a0.mimeData().hasUrls:
            a0.acceptProposedAction()
        else:
            super(MainWindow, self).dragEnterEvent(a0)

    def dropEvent(self, a0: QtGui.QDropEvent):
        mime = a0.mimeData()
        urls = mime.urls()
        err_msg = ""
        for url in urls:  # type: QtCore.QUrl
            fname = url.toLocalFile()
            try:
                dfs = loaders.load_file(fname)
            except Exception as inst:
                err_msg += "\nError: {:}.\ncould not load file: {:}".format(str(inst), fname)
                continue

            if isinstance(dfs, StructuredDataFrame):
                dfs = [dfs]
            for df in dfs:  # type: StructuredDataFrame
                try:
                    self.treeView_dataFrames.addDataFrame(df)
                except Exception as e:
                    err_msg += "\nError: {:}.\nSomething wrong could not load {:} as a StructuredDataFrame Object".format(str(e), df)
                    continue

        if err_msg.strip():
            fn.error_popup(err_msg)
        return

    def addNewVisualization(self, visualization: visualizations.base.Visualization):
        subwindow = VisualizationWindow(visualization())
        item = SubWindowListItem(subwindow, name=visualization.name)
        self.visList.addItem(item)
        self.mdiArea.addSubWindow(item.subWindow)
        item.showVisualization()

    def importDataFrameFromClipboard(self):
        df = StructuredDataFrame.from_clipboard()
        df.metadata["name"] = "DF - clipboard"
        self.treeView_dataFrames.addDataFrame(df)


def run(debug=False):
    fn.set_process_id("radie")
    # fn.setup_style()
    app = fn.instantiate_app(sys.argv)

    fn.set_popup_exceptions()
    cfg.set_dpi_scaling()
    main_window = MainWindow()
    main_window.show()

    if debug:
        from radie.plugins import examples
        main_window.treeView_dataFrames.addDataFrame(examples.example_powderdiffraction())
        main_window.treeView_dataFrames.addDataFrame(examples.example_vsm())

    # sys.exit(app.exec_())
    # use pyqtgraph.exit to avoid crashes on exit
    # https://stackoverflow.com/questions/27131294/error-qobjectstarttimer-qtimer-can-only-be-used-with-threads-started-with-qt
    app.exec_()
    pg.exit()


def debug():
    run(True)


if __name__ == "__main__":
    run(debug=False)



from collections import OrderedDict
import numpy as np
import pandas as pd
from radie.structures import StructuredDataFrame, register_data_structures


class TGA(StructuredDataFrame):
    """
    StructuredDataFrame for holding Thermogravimetric analysis data
    Default units are minutes, mg, and celsius
    Please convert results to default units during instantiation, no unit checking is done within
    """

    label = "Thermogravimetric Analysis"

    _required_metadata = StructuredDataFrame._required_metadata.copy()

    _x = "time" # minutes
    _y = "weight" # mg
    _z = "temperature" # celsius

    _required_columns = OrderedDict((
        ("temperature", float),
        ("weight", float),
        ("time", float),
    ))
    _column_properties = ["norm_weight", "deriv_norm_weight", "deriv_weight"]

    @property
    def norm_weight(self):
        """Normalize the weight using the first datapoint (samples can gain or lose mass)"""
        return self.weight/self.weight[0]

    @property
    def deriv_norm_weight(self):
        """
        Normalized weight derivative wrt temperature
        Returns
        -------
        Array of derivative of normalized weight wrt time
        """
        y = self.norm_weight.values
        x = self.temperature.values
        return pd.Series(np.gradient(y, x), name='Deriv. Weight (%/°C)')

    @property
    def deriv_weight(self):
        """
        Normalized weight derivative wrt temperature
        Returns
        -------
        Array of derivative of normalized weight wrt time
        """
        y = self.weight
        x = self.temperature
        return pd.Series(np.gradient(y, x), name='Deriv. Weight (mg/°C)')

register_data_structures(TGA)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'C:/Coding/Python/radie/radie/qt/ui\xyscatter.ui'
#
# Created by: PyQt5 UI code generator 5.9.2
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_XYScatter(object):
    def setupUi(self, XYScatter):
        XYScatter.setObjectName("XYScatter")
        XYScatter.resize(1000, 561)
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(XYScatter)
        self.verticalLayout_2.setContentsMargins(3, 3, 3, 3)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.splitter = QtWidgets.QSplitter(XYScatter)
        self.splitter.setOrientation(QtCore.Qt.Horizontal)
        self.splitter.setObjectName("splitter")
        self.layoutWidget = QtWidgets.QWidget(self.splitter)
        self.layoutWidget.setObjectName("layoutWidget")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.layoutWidget)
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout.setObjectName("verticalLayout")
        self.treeView_datasets = DFXYListView(self.layoutWidget)
        self.treeView_datasets.setObjectName("treeView_datasets")
        self.verticalLayout.addWidget(self.treeView_datasets)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.comboBox_lineStyle = QtWidgets.QComboBox(self.layoutWidget)
        self.comboBox_lineStyle.setObjectName("comboBox_lineStyle")
        self.comboBox_lineStyle.addItem("")
        self.comboBox_lineStyle.addItem("")
        self.horizontalLayout.addWidget(self.comboBox_lineStyle)
        self.label_alpha = QtWidgets.QLabel(self.layoutWidget)
        self.label_alpha.setObjectName("label_alpha")
        self.horizontalLayout.addWidget(self.label_alpha)
        self.slider_alpha = QtWidgets.QSlider(self.layoutWidget)
        self.slider_alpha.setMaximum(255)
        self.slider_alpha.setProperty("value", 255)
        self.slider_alpha.setOrientation(QtCore.Qt.Horizontal)
        self.slider_alpha.setObjectName("slider_alpha")
        self.horizontalLayout.addWidget(self.slider_alpha)
        self.label_alphaValue = QtWidgets.QLabel(self.layoutWidget)
        self.label_alphaValue.setObjectName("label_alphaValue")
        self.horizontalLayout.addWidget(self.label_alphaValue)
        self.verticalLayout.addLayout(self.horizontalLayout)
        self.groupBox_Legend = QtWidgets.QGroupBox(self.layoutWidget)
        self.groupBox_Legend.setCheckable(True)
        self.groupBox_Legend.setObjectName("groupBox_Legend")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout(self.groupBox_Legend)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.checkBox_legend_dfname = QtWidgets.QCheckBox(self.groupBox_Legend)
        self.checkBox_legend_dfname.setChecked(True)
        self.checkBox_legend_dfname.setObjectName("checkBox_legend_dfname")
        self.horizontalLayout_2.addWidget(self.checkBox_legend_dfname)
        self.checkBox_legend_label = QtWidgets.QCheckBox(self.groupBox_Legend)
        self.checkBox_legend_label.setChecked(False)
        self.checkBox_legend_label.setObjectName("checkBox_legend_label")
        self.horizontalLayout_2.addWidget(self.checkBox_legend_label)
        self.checkBox_legend_ycolumn = QtWidgets.QCheckBox(self.groupBox_Legend)
        self.checkBox_legend_ycolumn.setChecked(True)
        self.checkBox_legend_ycolumn.setObjectName("checkBox_legend_ycolumn")
        self.horizontalLayout_2.addWidget(self.checkBox_legend_ycolumn)
        self.verticalLayout.addWidget(self.groupBox_Legend)
        self.formLayout_plotOptions = QtWidgets.QFormLayout()
        self.formLayout_plotOptions.setObjectName("formLayout_plotOptions")
        self.label_xlabel = QtWidgets.QLabel(self.layoutWidget)
        self.label_xlabel.setObjectName("label_xlabel")
        self.formLayout_plotOptions.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.label_xlabel)
        self.lineEdit_xlabel = QtWidgets.QLineEdit(self.layoutWidget)
        self.lineEdit_xlabel.setObjectName("lineEdit_xlabel")
        self.formLayout_plotOptions.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.lineEdit_xlabel)
        self.label_ylabel = QtWidgets.QLabel(self.layoutWidget)
        self.label_ylabel.setObjectName("label_ylabel")
        self.formLayout_plotOptions.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.label_ylabel)
        self.lineEdit_ylabel = QtWidgets.QLineEdit(self.layoutWidget)
        self.lineEdit_ylabel.setObjectName("lineEdit_ylabel")
        self.formLayout_plotOptions.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.lineEdit_ylabel)
        self.verticalLayout.addLayout(self.formLayout_plotOptions)
        self.plotWidget = PlotWidget(self.splitter)
        self.plotWidget.setObjectName("plotWidget")
        self.verticalLayout_2.addWidget(self.splitter)

        self.retranslateUi(XYScatter)
        QtCore.QMetaObject.connectSlotsByName(XYScatter)

    def retranslateUi(self, XYScatter):
        _translate = QtCore.QCoreApplication.translate
        XYScatter.setWindowTitle(_translate("XYScatter", "DataFrames"))
        self.comboBox_lineStyle.setItemText(0, _translate("XYScatter", "line"))
        self.comboBox_lineStyle.setItemText(1, _translate("XYScatter", "points"))
        self.label_alpha.setText(_translate("XYScatter", "Scatter Alpha"))
        self.label_alphaValue.setText(_translate("XYScatter", "255"))
        self.groupBox_Legend.setTitle(_translate("XYScatter", "Legend"))
        self.checkBox_legend_dfname.setText(_translate("XYScatter", "DF Name"))
        self.checkBox_legend_label.setText(_translate("XYScatter", "Item Label"))
        self.checkBox_legend_ycolumn.setText(_translate("XYScatter", "Y-Column"))
        self.label_xlabel.setText(_translate("XYScatter", "X-Label"))
        self.label_ylabel.setText(_translate("XYScatter", "Y-Label"))

from radie.qt.plotlist import DFXYListView
from radie.qt.plotwidget import PlotWidget



import numpy as np
import struct

from radie import exceptions
from radie.loaders import Loader, register_loaders
from radie.plugins.structures.dsc import DSC
from radie.plugins.structures.tga import TGA

# Dictionary for unit conversion
# note that all keys are lower case
unit_conversion = {
    's': ('min', lambda x: x/60.),
    'sec': ('min', lambda x: x/60.),
    'seconds': ('min', lambda x: x/60.),
    'h': ('min', lambda x: x*60.),
    'hr': ('min', lambda x: x*60.),
    'hrs': ('min', lambda x: x*60.),
    'hours': ('min', lambda x: x*60.),
    'm': ('min', lambda x: x),
    'min': ('min', lambda x: x),
    'minutes': ('min', lambda x: x),
    'k': ('°C', lambda x: x+273.15),
    'f': ('°C', lambda x: (x-32)*5./9.),
    '°f': ('°C', lambda x: (x-32)*5./9.),
    'μg': ('mg', lambda x: x/1000.),
    'mg': ('mg', lambda x: x),
    'g': ('mg', lambda x: x*1000.),
    'kg': ('mg', lambda x: 1e6*x),
    'μw': ('mW', lambda x: 1000.*x),
    'mw': ('mW', lambda x: x),
    'w': ('mW', lambda x: x/1000.),
    'mj/°c': ('mJ/°C', lambda x: x)
    }


def convert_units(unit, x):
    """Attempts to convert to TGA structure defaults, otherwise returns input"""
    if unit.lower() in unit_conversion.keys():
        return unit_conversion[unit.lower()][0], unit_conversion[unit.lower()][1](x)
    else:
        return unit, x


def load_ta_instruments(fname, required_keys=None, required_kvs=None):
    """
    It appears most TA instruments raw files share a fairly common data structures
    This should return the header as a dictionary and the columns as a numpy array
    Unit conversion is attempted according to the module-level unit_converison dict

    Parameters
    ----------
    fname : basestring
        file path to TA instruments file
    required_keys: list
        list of strings that must be found in the header. "startswith" is used to check for existence
    required_kvs: dict
        dictionary of key-value pairs that should be found in the header, both
        "startswith" is used for both key and value. Useful ensuring the right loader function is used
         for teh right instrument

    Returns
    -------
    metadata : dict
        dictionary of strings corresponding to the header. There are three keys that have been added
        'columns', list of strings corresponding to column names without units (lower case)
        'units', list of strings corresponding to units for the columns (mixed case)
    results : numpy array
        array of floats, number of columns dictated by the input file
        TGA StructuredDataFrame
    """

    results = []
    with open(fname, "rb") as f:
        # Need to read by two or else utf-16 might return an error
        header_byte_string = b''
        while True:
            s = f.read(2)
            if s == b'\x0c\x00' or s == b'':
                break
            header_byte_string += s

        header_text = header_byte_string.decode('utf-16')
        lines = header_text.split('\r\n')

        if required_keys is not None:
            founds = [False] * len(required_keys)
            for i,required_key in enumerate(required_keys):
                for line in lines:
                    if line.startswith(required_key):
                        founds[i] = True
                        break

            if not all(founds):
                raise exceptions.IncorrectFileType

        if required_kvs is not None:
            founds = [False] * len(required_kvs)
            for i,(req_key, req_val) in enumerate(required_kvs.items()):
                for line in lines:
                    items = [x.strip() for x in line.split()]
                    if len(items) > 1:
                        key = items[0]
                        val = ' '.join(items[1:])
                        if key.startswith(req_key) and val.startswith(req_val):
                            founds[i] = True
                            break

            if not all(founds):
                raise exceptions.IncorrectFileType

        raw_column_headers = [s for s in lines if s.startswith('Sig')]

        # If no signals were found returns None
        if not raw_column_headers:
            raise exceptions.IncorrectFileType

        # After the b'\x0c\x00' there is a \x05 pad
        f.read(1)

        # the data is then floats corresponding to the number of signals
        read_format = 'f'*len(raw_column_headers)
        read_len = 4*len(raw_column_headers)

        while True:
            read_val = f.read(read_len)
            if read_val:
                if read_val.startswith(b'\x01\x04'):
                    # Reached "\x04" end of transmission marker
                    break
                else:
                    results.append(struct.unpack(read_format, read_val))
                    if results[-1][0] == -100.0:
                        # Data columns terminated by a -100 in the first column
                        results.pop(-1)
                        break
            else:
                # Reached EOF
                break

    if not results:
        raise exceptions.IncorrectFileType

    results = np.array(results)

    # Convert the header into metadata
    metadata = {}
    for line in lines:
        items = [x.strip() for x in line.split()]
        if len(items) > 1:
            key = items[0]
            val = ' '.join(items[1:])
            metadata[key] = val

    # Use sample as name
    metadata['name'] = metadata.get('Sample', '')

    # Extract column titles and units
    columns = []
    column_units = []
    if raw_column_headers:
        for i in range(len(raw_column_headers)):
            key = 'Sig{}'.format(i+1)
            head = metadata[key]
            columns.append(' '.join(head.split()[:-1]).lower())
            column_units.append(head.split()[-1][1:-1])

    # The TA Instruments default units seem to already be fine but convert just in case
    # Note that convert_units will return input if unable to convert
    for i, unit in enumerate(column_units):
        column_units[i], results[:,i] = convert_units(unit, results[:,i])
    metadata['columns'] = columns
    metadata['units'] = column_units

    return metadata, results


def load_dsc(fname):
    """

    Parameters
    ----------
    fname : file path

    Returns
    -------
    df_dsc : DSC
        DSC StructuredDataFrame
    """
    ta_out = load_ta_instruments(fname, required_kvs={"Instrument": "DSC Q2000"})
    if ta_out is None:
        raise exceptions.IncorrectFileType

    metadata, results = ta_out

    # Header typically reports the sample mass under the keyword "Size"
    size = metadata.get('Size')
    try:
        elements = size.split()
        metadata['mass'] = convert_units(elements[1], float(elements[0]))[1]
    except (AttributeError, ValueError, IndexError):
        metadata['mass'] = 1

    columns = metadata.pop('columns')
    underscored_columns = [s.lower().replace(' ', '_') for s in columns]
    df_dsc = DSC(data=results, columns=underscored_columns,
                 **metadata)
    return df_dsc


TA_Q2000_loader = Loader(load_dsc, DSC, [".001", ".002", ".003"], "TA Instruments Q2000")


def load_tga(fname):
    """
    Parameters
    ----------
    fname : file path

    Returns
    -------
    df_tga : TGA
        TGA StructuredDataFrame
    """
    ta_out = load_ta_instruments(fname, required_kvs={"Instrument": "TGA Q500"})
    if ta_out is None:
        raise exceptions.IncorrectFileType
    metadata, results = ta_out

    columns = metadata.pop('columns')
    underscored_columns = [s.lower().replace(' ', '_') for s in columns]

    df_tga = TGA(data=results,
                 columns=underscored_columns,
                 **metadata)
    return df_tga


TA_Q500_loader = Loader(load_tga, TGA, [".001", ".002", ".003"], "TA Instruments Q500")

register_loaders(
    TA_Q500_loader, TA_Q2000_loader,
)



"""plugins package, meant for custom data-structures and file-loaders and visualizations"""

import os
import pkgutil
import importlib

STRUCTURES_PKG_NAME = "structures"
STRUCTURES_PKG = __package__ + "." + STRUCTURES_PKG_NAME
LOADERS_PKG_NAME = "loaders"
LOADERS_PKG = __package__ + "." + LOADERS_PKG_NAME
VISUALIZATIONS_PKG_NAME = "visualizations"
VISUALZIATIONS_PKG = __package__ + "." + VISUALIZATIONS_PKG_NAME

this_dir = os.path.dirname(__file__)
structures_dir = os.path.join(this_dir, STRUCTURES_PKG_NAME)
loaders_dir = os.path.join(this_dir, LOADERS_PKG_NAME)
visualizations_dir = os.path.join(this_dir, VISUALIZATIONS_PKG_NAME)


disabled_plugins = (
    "powderdiffraction_siemensD500",
)


def import_structures():
    importlib.import_module(STRUCTURES_PKG)
    for finder, module_name, is_pkg in pkgutil.iter_modules([structures_dir]):
        if not is_pkg:
            importlib.import_module("." + module_name, STRUCTURES_PKG)


def import_loaders():
    importlib.import_module(LOADERS_PKG)
    for finder, module_name, is_pkg in pkgutil.iter_modules([loaders_dir]):
        if not is_pkg and module_name not in disabled_plugins:
            importlib.import_module("." + module_name, LOADERS_PKG)


def import_visualizations():
    importlib.import_module(VISUALZIATIONS_PKG)
    for finder, module_name, is_pkg in pkgutil.iter_modules([visualizations_dir]):
        if not is_pkg:
            importlib.import_module("." + module_name, VISUALZIATIONS_PKG)



from setuptools import setup, find_packages

# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

with open('radie/__init__.py') as fid:
    version_found = False
    for line in fid:
        if line.startswith('__version__'):
            version = line.split('=')[1].strip().replace("'", "")
            version_found = True
            break
    if not version_found:
        raise Exception('version string not found in __init__.py')

setup(
    name='radie',
    version=version,
    description='a python application to rapidly import and view experimental data',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://www.github.com/dvincentwest/radie',

    # Author details
    author='D. Vincent West',
    author_email='dvincentwest@gmail.com',

    # Choose your license
    license='GPLv2',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
        'Programming Language :: Python',
        'Topic :: Scientific/Engineering :: Visualization',
    ],
    keywords='data visualization',
    packages=find_packages(exclude=['contrib', 'docs', 'tests*']),
    python_requires='>=3.5',
    install_requires=[
        'numpy',
        'pandas',
        # 'PyQt5'  # left out because of naming conflicts with conda
        'pyqtgraph',
    ],
    extras_require={},
    package_data={
        'radie.plugins.examples': ['data/*'],
        'radie.qt': ['resources/icons/*.svg'],
        'radie.plugins.visualizations': ['icons/*.svg'],
    },
    data_files=[],
    entry_points={},
)



"""Customize the Basic PyQtGraph PlotWidget with styles and helpful functions"""

import types

import pyqtgraph as pg


class PlotWidget(pg.PlotWidget):
    def __init__(self, parent=None, background="default", **kwargs):
        super(PlotWidget, self).__init__(parent, background, **kwargs)

        # bolt on my PlotItem additions, for some reason all attempts to subclass don't really work
        self.plotItem.setupPlot = types.MethodType(PlotItem.setupPlot, self.plotItem)
        self.plotItem.resetLegend = types.MethodType(PlotItem.resetLegend, self.plotItem)
        self.plotItem.removeLegend = types.MethodType(PlotItem.removeLegend, self.plotItem)
        self.plotItem.setupPlot()


class PlotItem(pg.PlotItem):
    def __init__(self, *args, **kwargs):
        super(PlotItem, self).__init__(*args, **kwargs)
        self.setupPlot()

    def setupPlot(self):
        """
        A universal style for datauick plots

        Parameters
        ----------
        self : pg.PlotItem

        """
        self.addLegend()
        self.showAxis('top')
        self.showAxis('right')
        self.showGrid(True, True, 0.1)
        self.getAxis('top').setStyle(showValues=False)
        self.getAxis('right').setStyle(showValues=False)
        self.vb.setMouseMode(pg.ViewBox.RectMode)

    def removeLegend(self):
        """remove the legend"""
        legend = self.legend
        try:
            legend.scene().removeItem(legend)
            self.legend = None
        except AttributeError:
            pass

    def resetLegend(self):
        """
        clear out the old legend, add a new one and repopulate

        Parameters
        ----------
        self : pg.PlotItem

        Returns
        -------

        """
        self.removeLegend()
        self.addLegend()

        for item in self.listDataItems():
            if item.name() is not None:
                self.legend.addItem(item, item.name())



import warnings
from collections import OrderedDict

from . import structureddataframe
from .structureddataframe import StructuredDataFrame

__all__ = [
    StructuredDataFrame.__name__
]

structures = OrderedDict()
structures[StructuredDataFrame.__name__] = StructuredDataFrame


def register_data_structures(*sub_classes):
    """
    register StructuredDataFrame structures from the plugins folder

    Parameters
    ----------
    sub_classes
        StructuredDataFrame sub-classes as arguments

    """
    for cls in sub_classes:  # type: StructuredDataFrame
        if cls.__name__ in structures.keys():
            warnings.warn("overwriting {:s} df_class".format(cls.__name__))
        structures[cls.__name__] = cls
        globals()[cls.__name__] = cls  # put the class into this scope

        if cls.__name__ not in __all__:
            __all__.append(cls.__name__)


def print_available_structures():
    for key in structures.keys():
        print(key)



import os
import sys
import traceback
import csv
import warnings
import io
import json
import typing
from pathlib import Path

import pandas as pd

from radie import structures
from . import exceptions
from .structures.structureddataframe import StructuredDataFrame

loaders = dict()


class Loader(object):
    """object used to map file-extensions to functions that load StructuredDataFrame subclasses

    The loader object is a place holder that connects a list of file-extensions to a a function that will read those
    files and StructuredDataFrame subclass that the function will return.  Its main purpose is for book-keeping to allow
    automatic operations involving reading data files.  One key principle of operation is that the function will check
    and validate if the file is of the correct type and has all of the required necessary information.  If not, the
    function should raise a LoaderException, otherwise, the function returns an instance (or a list of
    instances) of the `StructuredDataFrame` subclass specified in the `cls` attribute

    Attributes
    ----------

    cls: typing.Type(StructuredDataFrame)
        the `StructuredDataFrame` sub-class that the loader function will load
    extensions : list, str
        a string or list of strings specifying the file types
    label : str
        the readable text that will represent this file loader
    module : str
        the module path of the loader function, helpful for debugging

    """

    def __init__(self, loader_function, cls, extensions, label):
        """
        Parameters
        ----------
        loader_function : function
            the function to load the StructuredDataFrame from a file, expects filename argument
        cls: typing.Type(StructuredDataFrame)
            the `StructuredDataFrame` sub-class that the loader function will load
        extensions : list, str
            a string or list of strings specifying the file types
        label : str
            the readable text that will represent this file loader
        """
        self._load = loader_function  # type: typing.Callable
        self.cls = cls

        if type(extensions) is list:
            self.extensions = extensions
        elif type(extensions) is str:
            self.extensions = [extensions]
        else:  # convert iterable type to list
            self.extensions = list(extensions)

        self.label = label

    @property
    def module(self):
        return self._load.__module__ + "." + self._load.__name__

    def load(self, filename):
        """load a `DataFrame` object from the loader function and return

        Parameters
        ----------
        filename : str

        Returns
        -------
        StructuredDataFrame

        """
        return self._load(filename)

    def __str__(self):
        return self.label + ": " + ", ".join(self.extensions)


def register_loaders(*loader_objects):
    """register Loader objects with the loaders.loaders book-keeping dict so we can keep track

    Parameters
    ----------
    loader_objects : Loader
        Loader objects supplies as arguments

    """
    for loader in loader_objects:  # type: Loader
        for ext in loader.extensions:
            if ext not in loaders.keys():
                loaders[ext] = []
            if loader in loaders[ext]:
                warnings.warn('overwriting Loader Object {:} for file extension {:}'.format(loader.label, ext))
            loaders[ext].append(loader)
        if loader.cls not in loader.cls.loaders():  # setup circular referencing of sorts
            loader.cls.add_loader(loader)


def from_clipboard(*args, **kwargs):
    df = pd.read_clipboard(*args, **kwargs)
    return StructuredDataFrame(df)


def load_csv(fname, name=None, encoding=None):
    """
    a function to read a generic csv file and return a dataframe. It is assumed that the file has the following
    structure:

        1. Metadata (Optional)
        2. DataBlock in CSV format with or without a header line

    The function goes to about 80% of the way through the file, assuming that that location is always within the
    datablock.  At this point, the Python csv module is used to "sniff" the datablock delimiter and also a number
    of columns and numeric signature are determined

    Next we go to the beginning of the file and look for the first sequence of 10 lines that are delimited by the
    sniffed delimiter, and has the same length and numeric signature of the datablock.  This marks the beginning of the
    datablock.

    Lastly we look at the line immediately above the datablock.  If that line has the same delimited length it is
    assumed to be the headerline.

    We use all of this information to pass the filename to pandas.read_csv, to read in the data-block and the headerline


    Parameters
    ----------
    fname : str or io.TextIOWrapper or io.StringIO

    Returns
    -------
    StructuredDataFrame

    """

    if not name:
        name = ''

    if isinstance(fname, Path):
        fname = str(fname)

    if type(fname) is str:
        if not os.path.isfile(fname):
            raise TypeError('fname must be a valid file, string blocks not yet supported')
        with open(fname, "r", encoding=encoding) as fid:
            if not name:
                name = os.path.basename(fname)
            data = io.StringIO(fid.read())
    elif type(fname) is io.TextIOWrapper:
        if not name:
            name = os.path.basename(fname.name)
        data = io.StringIO(fname.read)
    elif type(fname) is io.StringIO:
        data = fname
    else:
        raise TypeError('must provide a filename, a TextIOWrapper object, or a StringIO object')

    lines = data.readlines()
    data.seek(0)

    num_lines = len(lines)
    if num_lines > 100:
        # big file assume we find data_block at the 80% mark
        chunk_start = int(num_lines * 0.8)
        chunk_end = chunk_start + 10
    else:
        # small file, start at the end and work our way back to the first non-blank line
        for chunk_end in range(num_lines - 1, -1, -1):
            if lines[chunk_end] and not lines[chunk_end].isspace():
                break
        chunk_start = chunk_end - 10 if num_lines > 11 else 1

    chunk_lines = lines[chunk_start:chunk_end]
    chunk = "".join(chunk_lines)
    sniffer = csv.Sniffer()
    try:
        dialect = sniffer.sniff(chunk)
    except:
        err_msg = "\n".join(traceback.format_exception(*sys.exc_info()))
        raise exceptions.LoaderException('Sniffing delimieter failed\n{:s}'.format(err_msg))

    def numeric_signature(values):
        sig = list()
        for value in values:
            try:
                float(value)
                sig.append(True)
            except:
                sig.append(False)
        return sig

    lengths = [len(line.split(dialect.delimiter)) for line in chunk_lines]
    chunk_length = lengths[0]

    # if chunk lines are not all the same length, raise a LoaderException
    if not all([length == chunk_length for length in lengths]):
        raise exceptions.LoaderException('csv sample lines have different lengths')

    chunk_numeric_signature = numeric_signature(chunk_lines[0].split(dialect.delimiter))
    data_block_beginning = -1
    match_count = 0
    header_line_idx = None

    for i, line in enumerate(lines):
        line_vals = line.split(dialect.delimiter)

        if not len(line_vals) == chunk_length:
            match_count = 0
            continue

        line_numeric_signature = numeric_signature(line_vals)
        if not line_numeric_signature == chunk_numeric_signature:
            match_count = 0
            continue

        match_count += 1

        if match_count > 9:
            data_block_beginning = i - match_count + 1

            header_line_idx = data_block_beginning - 1
            if header_line_idx >= 0:
                header_line = lines[header_line_idx]
                header_line_vals = header_line.split(dialect.delimiter)

                if len(header_line_vals) == chunk_length + 1:  # assume we have a commented header
                    explicit_header = header_line_vals[1:]
                    header_line_idx = None;
                elif not len(header_line_vals) == chunk_length:
                    explicit_header = None
                    header_line_idx = None
            else:
                header_line_idx = None
            break

    if data_block_beginning < 0:
        raise exceptions.LoaderException('datablock not found')

    if header_line_idx is None:
        df = pd.read_csv(data, delimiter=dialect.delimiter, names=explicit_header,
                         skiprows=data_block_beginning, header=None, skip_blank_lines=False)
    else:
        df = pd.read_csv(data, delimiter=dialect.delimiter,
                         header=header_line_idx, skip_blank_lines=False)
    df = StructuredDataFrame(df.dropna(axis="index", how="all").dropna(axis='columns', how='all'), name=name)
    return df


def load_dftxt(fname):
    """load files from the `StructuredDataFrame.savetxt` method

    Parameters
    ----------
    fname : str

    Returns
    -------
    StructuredDataFrame
        fully specified datastructure from the dftxt file

    """
    header_block = []
    data_location = None
    data_sample = ''
    with open(fname, 'r') as fid:
        for line in iter(fid.readline, ''):
            if line.isspace():
                data_location = fid.tell()
            elif line.startswith('#'):
                header_block.append(line[1:])
                data_location = fid.tell()
            else:
                break

        if not data_location:
            raise exceptions.IncorrectFileType()
        try:
            meta_data = json.loads(''.join(header_block))  # type: dict
        except json.JSONDecodeError:
            raise exceptions.IncorrectFileType("header is not a valid json")
        except:
            raise exceptions.LoaderException("error when loading JSON")

        try:
            cls_key = meta_data.pop("class")
            file_type = meta_data.pop("file-type")
        except KeyError:
            raise exceptions.IncorrectFileType("df file missing required class for file-type information")
        else:
            try:
                cls = structures.structures[cls_key]  # type: typing.Type[StructuredDataFrame]
            except KeyError:
                raise exceptions.LoaderException("df class {:s} is not supported by any known DF Structure "
                                                 "classes".format(cls_key))

        if not cls.required_metadata().issubset(meta_data):
            raise exceptions.LoaderException("file meta-data does not match the required meta-data for the specified "
                                             "StructuredDataFrame Structure")

        for i, line in enumerate(iter(fid.readline, '')):
            data_sample += line
            if i == 10:
                break

        sniffer = csv.Sniffer()
        try:
            dialect = sniffer.sniff(data_sample)
        except:
            err_msg = "\n".join(traceback.format_exception(*sys.exc_info()))
            raise exceptions.LoaderException("Could not determine CSV block in datablock")

        fid.seek(data_location)
        df = pd.read_csv(fid, sep=dialect.delimiter)
        if not cls.required_columns().issubset(df.columns):
            raise exceptions.LoaderException(
                "StructuredDataFrame Columns do not match the required columns"
                "for the specified StructuredDataFrame Structure"
            )
        df = cls(df.dropna(axis="index", how="all").dropna(axis='columns', how='all'), **meta_data)
        return df


df_loader = Loader(load_dftxt, StructuredDataFrame, (".df"), "StructuredDataFrame text file")
register_loaders(df_loader)


def load_file(fname):
    """the catch-all convenience function to automatically load data-files

    This function is at the heart of one of the goals of Radie.  The idea is that any datafile that is supported
    by the radie structures and loaders can be passed to this function without needing to specify anything, and the
    proper Loader function will be determined automatically.  First a list of possible loader functions are found from
    the file extension.  They are then tried one by one, with each inappropriate function returning None, and continuing
    on to the next loader function.  The first time a valid StructuredDataFrame is returned, the function breaks the look and
    returns that object.  If no valid loader function is found, we attempt to read in the file using the load_csv
    automatic csv reader function.

    Parameters
    ----------
    fname : str
        the name of the datafile

    Returns
    -------
    StructuredDataFrame or list or tuple
        the output of the automatically determined loader function, could be a
        single StructuredDataFrame or a list of DataFrames

    """
    ext = os.path.splitext(fname)[1]

    try:
        loaders_ = loaders[ext]
    except KeyError:
        loaders_ = []

    for loader in loaders_:  # type: Loader
        try:
            dfs = loader.load(fname)
            if isinstance(dfs, StructuredDataFrame):
                print("successfully loaded file as {:}".format(type(dfs).__name__))
                return dfs
            elif type(dfs) in (list, tuple):
                df_list = [df for df in dfs if isinstance(df, StructuredDataFrame)]
                if not df_list:
                    raise exceptions.LoaderException("function {:s} did not return any StructuredDataFrame "
                                                     "objects from file {:s}".format(loader.module, fname))
                print("loaded {:d} dataframes from {:}".format(len(df_list), os.path.basename(fname)))
                return df_list
            else:
                raise exceptions.LoaderException("function {:s} did not return any StructuredDataFrame "
                                                 "objects from file {:s}".format(loader.module, fname))
        except exceptions.IncorrectFileType:
            continue
        except exceptions.LoaderException as loader_exception:
            raise loader_exception
        except Exception:
            err_msg = "loader function {:}.{:} failed".format(loader._load.__module__, loader._load.__name__)
            err_msg += "\n" + "\n".join(traceback.format_exception(*sys.exc_info()))
            raise Exception(err_msg)

    # no suitable loader was found, try the universal "load_csv" function
    try:
        df = load_csv(fname)
        return df
    except ValueError:  # in case of a unicode file not loaded properly
        try:
            df = load_csv(fname, encoding="utf-8")
            return df
        except exceptions.LoaderException:
            raise exceptions.LoaderNotFound("No loader found for the specified file")
        except Exception as e:
            raise e
    except exceptions.LoaderException:
        raise exceptions.LoaderNotFound("No loader found for the specified file")
    except Exception:
        raise exceptions.LoaderNotFound("load_csv failed for unexpected reason\n{:s}".format("\n".join(
            traceback.format_exception(*sys.exc_info()))))



from pyqtgraph import PlotItem
from pyqtgraph.exporters.Exporter import Exporter
import pyqtgraph.functions as fn
import win32com.client as win32


def irgb_to_long(r, g, b):
    return 65536 * int(b) + 256 * int(g) + int(r)


def qcolor_to_long(qcolor):
    r = qcolor.red()
    g = qcolor.green()
    b = qcolor.blue()
    return irgb_to_long(r, g, b)


def array_to_list(xs):
    """recursively convert an array to a list."""
    try:
        return [array_to_list(x) for x in xs]
    except TypeError:
        return xs


def set_array_range(worksheet, itop_row, ileft_col, data):
    """insert a 1d or 2d array starting at given location in provided worksheet.

    Args:
        worksheet: win32 Excel object
        itop_row: top row index (zero based)
        ileft_col: left column index (zero based)
        data: numpy array

    Returns:
        win32 Excel Range object
    """
    if len(data.shape) == 1:
        data = data.reshape(data.shape[0], 1)
    ibottom_row = itop_row + data.shape[0] - 1
    iright_col = ileft_col + data.shape[1] - 1
    xlrange = worksheet.Range(
        worksheet.Cells(itop_row + 1, ileft_col + 1),
        worksheet.Cells(ibottom_row + 1, iright_col + 1))

    xlrange.Value = data.tolist()
    return xlrange


def set_cell(worksheet, i, j, val):
    """Set a cell value

    Parameters
    ----------
    worksheet : Worksheet object
    i : integer
        zero based row index
    j : integer
        zero based column index
    val : (unspecified)
        value to set in the excel cell

    Return
    ------
    range : Range object
        range including the cell that was set
    """
    range_str = "%s" % gen_excel_name(i, j)
    xlrange = worksheet.Range(range_str)
    xlrange.Value = val
    return xlrange


def gen_excel_name(row, col):
    """Translate (0,0) into A1.

    Args:
        row: row index (zero-based)
        col: column index (zero-based)

    Returns:
        string representing cell name

    """

    if col < 26:
        col_name = chr(col + ord('A'))
    else:
        col_name = chr(int(col / 26) - 1 + ord('A')) + chr((col % 26) + ord('A'))

    return "%s%s" % (col_name, row + 1)


class VariableContainer:
    pass


class ExcelExporter(Exporter):
    Name = "Excel Chart"
    enums = VariableContainer()
    enums.xlXYScatter = -4169
    enums.xlXYScatterLines = 74
    enums.xlXYScatterLinesNoMarkers = 75
    enums.xlXYScatterSmooth = 72
    enums.xlXYScatterSmoothNoMarkers = 73
    enums.xlLow = -4134
    enums.xlNextToAxis = 4
    enums.xlHigh = -4127
    enums.xlRight = -4152
    enums.msoTrue = -1
    enums.msoElementPrimaryCategoryAxisTitleAdjacentToAxis = 301
    enums.msoElementPrimaryValueAxisTitleRotated = 309
    enums.xlCategory = 1
    enums.xlPrimary = 1
    enums.xlValue = 2
    enums.xlLineStyleNone = -4142

    symbol_map = {'o': 8,
                  's': 1,
                  't': 3,
                  'd': 2,
                  '+': 9}

    def __init__(self, item):
        Exporter.__init__(self, item)
        # Indices to track data columns starting points in data worksheet
        self.data_icol = 0
        self.data_irow = 1
        self.chart = None
        self.worksheet = None
        self.workbook = None

    def parameters(self):
        return None

    def export(self, fileName=None):
        """
        If *fileName* is None, pop-up a file dialog.
        If *toBytes* is True, return a bytes object rather than writing to file.
        If *copy* is True, export to the copy buffer rather than writing to file.
        """
        if not isinstance(self.item, PlotItem):
            raise Exception("Matplotlib export currently only works with plot items")

        # acquire application object, which may start application
        application = win32.gencache.EnsureDispatch("Excel.Application")
        try:
            version = float(application.Version)
        except ValueError:
            version = 12.0  # defaults to Excel 2007
        # create new file ('Workbook' in Excel-vocabulary)
        self.workbook = application.Workbooks.Add()
        # store default worksheet object so we can delete it later
        default_worksheet = self.workbook.Worksheets(1)

        # build new chart (on separate page in workbook)
        self.chart = self.workbook.Charts.Add()

        self.chart.ChartType = self.enums.xlXYScatterLinesNoMarkers
        self.chart.Name = "Plot"
        if version >= 12:
            self.chart.ApplyLayout(1)
            self.chart.ChartArea.Format.Line.Visible = False  # not sure if this is needed
            self.chart.ChartArea.Border.LineStyle = self.enums.xlLineStyleNone
        else:
            self.chart.PlotArea.Interior.Color = 16777215
            self.chart.HasTitle = True
            self.chart.Axes(self.enums.xlCategory, self.enums.xlPrimary).HasTitle = True
            self.chart.Axes(self.enums.xlCategory, self.enums.xlPrimary).AxisTitle.Characters.Text = "xaxis"
            self.chart.Axes(self.enums.xlValue, self.enums.xlPrimary).HasTitle = True
            self.chart.Axes(self.enums.xlValue, self.enums.xlPrimary).AxisTitle.Characters.Text = "yaxis"
            self.chart.ChartArea.Border.LineStyle = self.enums.xlLineStyleNone

        # create data worksheet
        self.worksheet = self.workbook.Worksheets.Add()
        self.worksheet.Name = "Plot data"

        for item in self.item.curves:
            self.add_curve_to_chart(item)

        # setup axes
        x_axis = self.chart.Axes().Item(self.enums.xlCategory, self.enums.xlPrimary)
        y_axis = self.chart.Axes().Item(self.enums.xlValue, self.enums.xlPrimary)
        x_axis.HasMajorGridlines = True
        y_axis.HasMajorGridlines = True

        # setup axis labels
        x_label = self.item.axes['bottom']['item'].label.toPlainText()
        y_label = self.item.axes['left']['item'].label.toPlainText()
        x_axis.AxisTitle.Text = x_label if x_label else "x axis"
        y_axis.AxisTitle.Text = y_label if y_label else "y axis"
        x_axis.AxisTitle.Font.Size = 20
        y_axis.AxisTitle.Font.Size = 20
        x_axis.TickLabels.Font.Size = 18
        y_axis.TickLabels.Font.Size = 18

        # Format and place legend
        self.chart.Legend.Font.Size = 18
        self.chart.Legend.IncludeInLayout = False
        self.chart.Legend.Position = self.enums.xlRight
        self.chart.Legend.Format.Fill.Visible = self.enums.msoTrue
        self.chart.Legend.Format.Fill.ForeColor.RGB = irgb_to_long(255, 255, 255)
        self.chart.Legend.Format.Fill.Transparency = 0
        self.chart.Legend.Format.Line.Visible = self.enums.msoTrue
        self.chart.Legend.Format.Line.ForeColor.RGB = irgb_to_long(115, 115, 115)
        self.chart.Legend.Format.Line.Transparency = 0

        # Set the labels position for xaxis
        x_axis.TickLabelPosition = self.enums.xlLow

        if self.item.titleLabel.text:
            self.chart.ChartTitle.Text = self.item.titleLabel.text
        else:
            self.chart.ChartTitle.Delete()

        # Set scale
        scale = self.item.viewRange()

        # if False:
        #     x_axis.MinimumScale = scale[0][0]
        #     x_axis.MaximumScale = scale[0][1]
        #     y_axis.MinimumScale = scale[1][0]
        #     y_axis.MaximumScale = scale[1][1]

        # remove default worksheet
        default_worksheet.Delete()

        # make stuff visible now.
        self.chart.Activate()
        application.Visible = True

    def add_xy_to_worksheet(self, x, y, x_header=None, y_header=None):
        set_cell(self.worksheet, 0, self.data_icol, x_header)
        set_cell(self.worksheet, 0, self.data_icol + 1, y_header)
        x_column = set_array_range(self.worksheet, self.data_irow, self.data_icol, x)
        y_column = set_array_range(self.worksheet, self.data_irow, self.data_icol + 1, y)
        self.data_icol += 2
        return x_column, y_column

    def add_curve_to_chart(self, item):
        """
        Adds the item to teh chart and corresponding data to the worksheet

        Parameters
        ----------
        item: pyqtgraph CurveItem

        Returns
        -------

        """

        # Get data
        x, y = item.getData()
        # Get headers
        name = None
        if hasattr(item, 'implements') and item.implements('plotData') and item.name() is not None:
            name = item.name()
            x_header = "{}_x".format(name)
            y_header = "{}_y".format(name)
        else:
            x_header = "{:04d}_x".format(int(self.data_icol / 2))
            y_header = "{:04d}_x".format(int(self.data_icol / 2))

        # Write to data worksheet
        x_column, y_column = self.add_xy_to_worksheet(x, y, x_header=x_header, y_header=y_header)

        # Add series to Excel Chart
        series = self.chart.SeriesCollection().NewSeries()
        series.XValues = x_column
        series.Values = y_column
        if name is not None:
            series.Name = name

        pen = item.opts["pen"]
        symbol = item.opts["symbol"]

        if pen is not None:
            line_color = qcolor_to_long(fn.mkPen(pen).color())
            series.Border.Color = line_color
            series.Format.Line.ForeColor.RGB = line_color
            series.Format.Fill.BackColor.RGB = line_color

        if symbol is not None:
            series.MarkerStyle = self.symbol_map.get(symbol, 1)
            series.MarkerSize = item.opts["symbolSize"]
            series.MarkerBackgroundColor = qcolor_to_long(fn.mkPen(item.opts["symbolPen"]).color())
            series.MarkerForegroundColor = qcolor_to_long(fn.mkBrush(item.opts["symbolBrush"]).color())

        if name is None:
            last_index = self.chart.Legend.LegendEntries().Count
            self.chart.Legend.LegendEntries(last_index).Delete()


ExcelExporter.register()



"""define the generic Visualization class"""
import os

from PyQt5 import QtWidgets, QtGui
from ...structures.structureddataframe import StructuredDataFrame


class Visualization(QtWidgets.QWidget):
    """base class for QtWidget visualizations

    Very minimal skeleton defining the expected behavior a QWidget used to visualize a StructuredDataFrame.  Almost
    anything that will fit into a QtWidget is acceptable as a Visualization.  Most of the internals are defined in
    subclass.

    Attributes
    ----------
    name : str
        *cls var*, The name of the visualization to be displayed in menus
    description : str
        *cls var*, a short one line description
    supported_classes : list of Type(StructuredDataFrame)
        *cls var*, provide an easy mechanism to validate if the StructuredDataFrame objects in the visualization
        will play nice with the visualization

    Notes
    -----
    Private class variables for subclassing:

    _icon_image : str
        the path to the icon image used for this visualization
    _icon : QtGui.QIcon
        the Qt Icon constructed using `_icon_image` after an icon is requested from the application

    """

    name = 'Base'  # type: str
    description = None  # type: str
    _icon_image = None  # type: str
    _icon = None  # type: QtGui.QIcon

    supportedClasses = (  # must be a tuple
        StructuredDataFrame,
    )

    @classmethod
    def icon(cls):
        """return the icon specified in cls._icon_image, or a blank icon if no icon is specified

        Returns
        -------
        QtGui.QIcon

        Notes
        -----
        Certain versions of PyQt fail if a QIcon is constructed before the QApplication.  We specify the icon
        image file in cls._icon_image and then a QIcon is then constructed the first time this method is called.
        Subsequent calls return the previously constructed QIcon object

        """
        if isinstance(cls._icon, QtGui.QIcon):
            return cls._icon
        elif not cls._icon_image:
            return QtGui.QIcon()

        if os.path.isfile(cls._icon_image):
            try:
                cls._icon = QtGui.QIcon(cls._icon_image)
                return cls._icon
            except:
                cls._icon = None

        return QtGui.QIcon()

    def exportToExcel(self):
        """optional method to export the visualization to an excel file"""
        raise NotImplementedError

    def copyImage(self):
        """put an image of the visualization on the clipboard"""
        raise NotImplementedError



"""Loader for csv output of Horiba LA-960 particle size analyzer"""
import numpy as np
import re

from radie import exceptions
from radie.loaders import Loader, register_loaders
from radie.plugins.structures.psd import PSD
from radie.util import iso_date_string


class Parameter(dict):
    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)
        self.reset_defaults(**kwargs)

    def reset_defaults(self, **kwargs):
        d = {'value': None,
             'type': str,
             'key': None,
             'unit': None,
             'delimiter': None,
             'raw_line': None,
             'raw_value': None,
             'format': '',
             'padding': True,
             'merge_consecutive_delimiters': True,
             'value_has_whitespace': True}
        self.update(d)
        self.update(**kwargs)

    def __str__(self):
        if self.get('value') is not None:
            if self.get('unit') is not None:
                return '{key}:{value:{format}} {unit}'.format(**self)
            else:
                return '{key}:{value:{format}}'.format(**self)

    def read_line(self, line):
        self['raw_value'] = line
        self['value'] = self['type'](line)
        return self['value']


class KeyValParameter(Parameter):
    def __init__(self, key=None, **kwargs):
        Parameter.__init__(self, **kwargs)
        self['key'] = key
        if self.get('raw_line') is not None:
            self.read_line(self['raw_line'])

    def read_line(self, line):
        if self.check_line(line):
            try:
                items = line.split(self['delimiter'])
                self['raw_value'] = self['delimiter'].join(items[1:])
                return self.update_value()
            except (TypeError, ValueError, AttributeError):
                self['raw_value'] = None
                self['value'] = None

    def update_value(self):
        valstr = self['raw_value'].strip()
        if self.get('unit', False):
            if self['unit'] in valstr:
                iunit = valstr.index(self['unit'])
                valstr = valstr[:iunit]
        try:
            self['value'] = self['type'](valstr)
        except (ValueError, AttributeError):
            self['value'] = None
        return self['value']

    def check_line(self, line):
        if self.get('key') is not None:
            return line.strip().startswith(self['key'])
        else:
            return False


class RegexParameter(Parameter):
    def __init__(self, regex=None, **kwargs):
        Parameter.__init__(self, **kwargs)
        if regex is None:
            self.gen_regex()
        else:
            self['regex'] = regex
        if self.get('raw_line') is not None:
            self.read_line(self['raw_line'])

    def gen_regex(self):
        # pdb.set_trace()
        key = re.escape(self.get('key')) if self.get('key') else ''
        if self.get('delimiter') is None:
            delimiter = r'\s+'
        else:
            delimiter = '({0})'.format(re.escape(self['delimiter']))
            if self.get('merge_consecutive_delimiters', False):
                delimiter += '+'
        unit = re.escape(self['unit']) if self.get('unit') is not None else ''
        padding = r'\s*?' if self.get('padding') is not None else ''
        if self.get('value_has_whitespace', False):
            raw_value = r'(?P<raw_value>.+)'
        else:
            raw_value = r'(?P<raw_value>\w+)'

        regex = r'{key}{padding}{delimiter}{padding}{raw_value}.*{unit}'.format(key=key, delimiter=delimiter,
                                                                                 unit=unit, padding=padding,
                                                                                 raw_value=raw_value)
        self['regex'] = regex
        return regex

    def read_line(self, line):
        self['raw_line'] = line
        self['m'] = re.compile(self['regex'])
        mo = self['m'].search(line.strip())
        if mo is not None:
            self.update(mo.groupdict())
            self.update_value()
        else:
            self['value'] = None
        return self['value']

    def update_value(self):
        try:
            self['value'] = self['type'](self['raw_value'].strip())
        except ValueError:
            self['value'] = None
        return self['value']


def load_csv(fname):
    """

    Parameters
    ----------
    fname : file path

    Returns
    -------
    df_psd : PSD
        PSD StructuredDataFrame
    """

    starting_lines = [
    "Median size",
    "Mean size",
    "Variance",
    "St. Dev.",
    "Mode size",
    "Span",
    "Geo. mean size",
    "Geo. variance",
    "Diameter on cumulative",
    "D10",
    "D90",
    "D(v,0.1)",
    "D(v,0.5)",
    "D(v,0.9)",]

    # Open while checking for expected format,
    # Want to fail as fast as possible
    lines = []
    with open(fname, "r") as reader:
        for i, line in enumerate(reader):
            if i < len(starting_lines):
                if not line.startswith(starting_lines[i]):
                    raise exceptions.IncorrectFileType()
            lines.append(line)

    # The Horiba can output with a user defined delimiter character
    # commas and tabs are currently supported
    if ',' in lines[0]:
        delimiter = ','
    elif '\t' in lines[0]:
        delimiter = '\t'
    else:
        raise exceptions.IncorrectFileType("only commas and tabs are supported as the delimiter")

    metadata = {}
    parameters = {}
    parameters['sample'] = RegexParameter(key='Sample Name', delimiter=delimiter)
    parameters['lot'] = RegexParameter(key='Lot Number', delimiter=delimiter)

    lines = [line.strip() for line in lines]

    while not lines[0].startswith('Diameter (\xb5m)'):
        line = lines.pop(0)
        for k, p in parameters.items():
            if line.startswith(p['key']):
                try:
                    value = p.read_line(line)
                    if value is not None:
                        break
                except:
                    pass

    if parameters['sample']['value'] is None:
        parameters['sample']['value'] = 'PSD_{}'.format(iso_date_string())

    for k, p in parameters.items():
        metadata[k] = p['value']
    metadata['name'] = parameters['sample']['value']

    lines.pop(0)
    # The last two lines are blank/null values
    data = np.array([[float(x) for x in line.split(delimiter)] for line in lines[:-2]])

    df_psd = PSD(data=data[:,:2],
                 columns=['diameter','frequency'],
                 **metadata)
    return df_psd

LA960_csv_loader = Loader(load_csv, PSD, [".csv"], "Horiba LA-960")

register_loaders(
    LA960_csv_loader,
)





"""plot a histogram of a single pandas series"""

import os

import numpy as np
from PyQt5 import QtCore, QtWidgets
import pyqtgraph as pg

from ...structures.structureddataframe import StructuredDataFrame
from .. import colors, cfg, dpi
from .. import plotwidget
from .. import plotlist
from .base import Visualization


class DFItem(plotlist.DFItem):
    """an item class with pyqtgraph xy curve handles of type PlotDataItem"""
    def __init__(self, ref, item_list, name=None):
        super(DFItem, self).__init__(ref, item_list, name)
        self.plotDataItem = pg.PlotDataItem()
        self.color = None
        self.density = None
        self.bins = None

    def setText(self, value):
        self.text = value
        self.plotDataItem.setData(name=value)
        self.plotDataItem.updateItems()

    def calculateHistogram(self, bins, density):
        """calculate the histogram values, given the number of bins

        Parameters
        ----------
        bins : int
        density : bool

        """
        self.density, self.bins = np.histogram(self.x_data(), bins, density=density)


class Histogram(Visualization):
    """A generic class providing the visualization of a column of data as a binned histogram

    This visualization is intended for use where we have populations of data and we want to compare the histgrams
    of different variables in relation to eachother.  The list widget provides a single combobox to select which
    column of data will be binned and visualized

    """

    name = "Histogram"
    description = "A generic visualization of histograms for a Series of values"
    _icon_image = os.path.join(cfg.icon_path, "histogramicon.svg")

    def __init__(self, parent=None):
        super(Histogram, self).__init__(parent)

        # --- setupUi --- #
        self.resize(1000, 600)
        self.setWindowTitle("Histogram")
        self.vlay_main = QtWidgets.QVBoxLayout(self)
        self.vlay_main.setContentsMargins(3, 3, 3, 3)
        self.splitter = QtWidgets.QSplitter(self)
        self.splitter.setOrientation(QtCore.Qt.Horizontal)
        self.dataListWidget = QtWidgets.QWidget(self.splitter)
        self.vlay_dataListWidget = QtWidgets.QVBoxLayout(self.dataListWidget)
        self.vlay_dataListWidget.setContentsMargins(0, 0, 0, 0)
        self.treeView_datasets = plotlist.DFSeriesListView(self.dataListWidget)
        self.vlay_dataListWidget.addWidget(self.treeView_datasets)
        self.hlay_parameters = QtWidgets.QHBoxLayout()
        self.checkBox_plotDensity = QtWidgets.QCheckBox("Probability Density")
        self.checkBox_plotDensity.setCheckState(QtCore.Qt.Checked)
        self.hlay_parameters.addWidget(self.checkBox_plotDensity)
        self.spinBox_bins = QtWidgets.QSpinBox()
        self.spinBox_bins.setValue(10)
        self.spinBox_bins.setMaximum(50)
        self.spinBox_bins.setMinimum(3)
        self.hlay_parameters.addWidget(self.spinBox_bins)
        self.vlay_dataListWidget.addLayout(self.hlay_parameters)
        self.plotWidget = plotwidget.PlotWidget(self.splitter)
        self.vlay_main.addWidget(self.splitter)

        self.splitter.setStretchFactor(0, 0)
        self.splitter.setStretchFactor(1, 1)
        self.splitter.setSizes(dpi.width_by_height(280, 720))
        # --- end setupUi --- #

        self.treeView_datasets.setItemClass(DFItem)
        self.treeView_datasets.supportedClasses = self.supportedClasses

        self.treeView_datasets.model().itemsAdded.connect(self.addCurves)
        self.treeView_datasets.model().itemAccessorChanged.connect(self.itemDataChanged)
        self.treeView_datasets.model().itemsDeleted.connect(self.processNewLayout)
        self.treeView_datasets.model().itemToggled.connect(self.itemToggled)
        self.treeView_datasets.model().itemTextUpdated.connect(self.plotWidget.plotItem.resetLegend)
        self.checkBox_plotDensity.stateChanged.connect(self.recalculateHistograms)
        self.spinBox_bins.editingFinished.connect(self.recalculateHistograms)

        self._colors = None
        self.resetColors()

    def nextColor(self):
        return next(self._colors)

    def resetColors(self):
        self._colors = colors.colors()

    def recalculateHistograms(self):
        for item in self.treeView_datasets.iterItems():
            item.calculateHistogram(self.spinBox_bins.value(), self.checkBox_plotDensity.isChecked())
            item.plotDataItem.setData(x=item.bins, y=item.density, stepMode=True)

    def itemDataChanged(self, item: DFItem):
        if not item.isChecked():
            return
        item.calculateHistogram(self.spinBox_bins.value(), self.checkBox_plotDensity.isChecked())
        item.plotDataItem.setData(x=item.bins, y=item.density, stepMode=True)

    def processNewLayout(self):
        self.plotWidget.plotItem.clear()
        self.plotWidget.plotItem.resetLegend()
        for item in self.treeView_datasets.iterItems():
            self.plotWidget.addItem(item.plotDataItem)

    def itemToggled(self, item):
        """
        process checking/unchecking of a StructuredDataFrame in the plot

        Parameters
        ----------
        item : DFItem
        """
        if item.checkState:
            item.plotDataItem.setData(item.bins, item.density, stepMode=True, pen=item.color, name=item.text)
        else:
            item.plotDataItem.setData(name=None, stepMode=False)
            item.plotDataItem.clear()
        self.plotWidget.plotItem.resetLegend()

    def addCurves(self, items):
        """
        Parameters
        ----------
        items : list of DFItem
        """
        for item in items:  # type: DFItem
            item.color = self.nextColor()
            item.calculateHistogram(self.spinBox_bins.value(), density=self.checkBox_plotDensity.isChecked())
            item.plotDataItem.setData(item.bins, item.density, stepMode=True, pen=item.color, name=item.text)
            self.plotWidget.plotItem.addItem(item.plotDataItem)


def test():
    import sys
    from ..masterdftree import DFReference
    from .. import cfg
    from .. import functions as fn
    app = fn.instantiate_app()
    fn.reset_excepthook()
    cfg.set_dpi_scaling()

    rng = np.random.RandomState(10)  # deterministic random data
    a = np.hstack((rng.normal(size=1000), rng.normal(loc=5, scale=2, size=1000)))
    b = np.hstack((rng.normal(size=1000, scale=2), rng.normal(loc=5, scale=1, size=1000)))
    data = np.vstack((a, b)).T
    df = StructuredDataFrame(data, columns=("a", "b"), name="Deterministic Random Data")
    ref1 = DFReference(df, None)

    plot = Histogram()
    plot.treeView_datasets.addDataFrames(ref1, ref1)
    plot.show()

    sys.exit(app.exec_())



try:
    import PyQt5
except ImportError:
    raise Exception("PyQt5 module note found, run `conda install pyqt` or `pip install PyQt5`")

from . import cfg
from .functions import *

from .. import plugins

plugins.import_visualizations()



import numpy as np
import os.path

from radie.loaders import Loader, register_loaders
from radie.plugins.structures.powderdiffraction import PowderDiffraction, CuKa


def convert_to_line_peaks(xy, ybase=0, ynorm=100.):
    line_peaks = []
    for x, y in xy:
        line_peaks.append([x, ybase])
        line_peaks.append([x, y])
        line_peaks.append([x, ybase])

    # Normalize peaks
    line_peaks = np.array(line_peaks)
    line_peaks[:, 1] = line_peaks[:, 1] / line_peaks[:, 1].max() * ynorm
    return line_peaks


def load_peaks(fname, name=None):
    """
    .peaks file
    Expected format is just "#" commented header and two columns of 2theta and intensities for the peaks
    Data gets reworked to make bars in a line graph with intensities of max 1000

    Parameters
    ----------
    fname : str
        filename
    name : str
        measurement identifier

    Returns
    -------
    df_xrd : single PowderDiffraction

    """
    ybase = 0.0
    ynorm = 100.

    with open(fname, 'rb') as f:
        peaks = np.loadtxt(f, delimiter=",")

    # The expectation is that this is a list of 2theta and peaks, in order to make it a line
    # it should be processed and shoulder point added

    line_peaks = convert_to_line_peaks(peaks, ybase=ybase, ynorm=ynorm)

    # Use filename for name
    name = os.path.splitext(os.path.basename(fname))[0]
    df_xrd = PowderDiffraction(data=line_peaks,
                               columns=['twotheta', 'intensity'],
                               wavelength=CuKa,
                               name=name)

    return df_xrd


def calc_twotheta_from_d(d, wavelength=CuKa):
    """
    Calcuate twotheta at a given wavelength from the length of the reciprocal lattice vector `Q`

    Parameters
    ----------
    d : np.ndarray, float
        d spacings
    wavelength : float
        the desired wavelength in angstroms, defaults to CuKa

    Returns
    -------
    twotheta : np.ndarray, float
        diffraction angle in degrees

    """
    return np.rad2deg(2 * np.arcsin(wavelength / (2 * d)))


def load_dpeaks(fname, name=None):
    """
    .peaks file
    Expected format is just "#" commented header and two columns of 2theta and intensities for the peaks
    Data gets reworked to make bars in a line graph with intensities of max 1000

    Parameters
    ----------
    fname : str
        filename
    name : str
        measurement identifier

    Returns
    -------
    df_xrd : single PowderDiffraction

    """
    ybase = 0.0
    ynorm = 100.

    with open(fname, 'rb') as f:
        peaks = np.loadtxt(f, delimiter=",")

    # The expectation is that this is a list of d spacings
    peaks[:, 0] = calc_twotheta_from_d(peaks[:, 0])
    line_peaks = convert_to_line_peaks(peaks)

    # Use filename for name
    name = os.path.splitext(os.path.basename(fname))[0]
    df_xrd = PowderDiffraction(data=line_peaks,
                               columns=['twotheta', 'intensity'],
                               wavelength=CuKa,
                               name=name)

    return df_xrd


xrd_peaks_loader = Loader(load_peaks, PowderDiffraction, [".peaks"], "twotheta peak file")
xrd_dpeaks_loader = Loader(load_dpeaks, PowderDiffraction, [".dpeaks"], "d-spacing peak file")
register_loaders(xrd_peaks_loader, xrd_dpeaks_loader)



"""A convenience module for getting dimensions in the context of the dpi of the monitor"""

from . import cfg


def scaling(dpi=None):
    """
    get the scaling factor based on the dpi of the current screen

    Parameters
    ----------
    dpi : int or float

    Returns
    -------
    float

    """
    if dpi:
        return dpi / cfg.dpi_100percent
    else:
        return cfg.dpi_scaling


def length(pixels, dpi=None):
    return int(pixels * scaling(dpi))


def width_by_height(w, h, dpi=None):
    scale = scaling(dpi)
    return int(w * scale), int(h * scale)
